#
# Localized Starting Techs
# CvLocalizedStartingTechsEventManager
# 

from CvPythonExtensions import *

import Popup as PyPopup

import CvUtil

import CvEventManager
import sys
import PyHelpers
import CvConfigParser
import math
import time
import CvScreensInterface
			
gc = CyGlobalContext()	

PyPlayer = PyHelpers.PyPlayer
PyGame = PyHelpers.PyGame()
localText = CyTranslator()

TECH_FISHING = "TECH_FISHING"
TECH_AGRICULTURE = "TECH_AGRICULTURE"
TECH_THE_WHEEL = "TECH_THE_WHEEL"
TECH_MYSTICISM = "TECH_MYSTICISM"
TECH_HUNTING = "TECH_HUNTING"
TECH_MINING = "TECH_MINING"

PLOT_HILLS = PlotTypes.PLOT_HILLS
PLOT_LAND = PlotTypes.PLOT_LAND
PLOT_OCEAN = PlotTypes.PLOT_OCEAN
PLOT_PEAK = PlotTypes.PLOT_PEAK

g_plotMap = { 
				"PLOT_HILLS" : PLOT_HILLS,
				"PLOT_LAND" : PLOT_LAND,
				"PLOT_OCEAN" : PLOT_OCEAN,
				"PLOT_PEAK" : PLOT_PEAK,
			}

TERRAIN_GRASS = "TERRAIN_GRASS"
TERRAIN_PLAINS = "TERRAIN_PLAINS"
TERRAIN_DESERT = "TERRAIN_DESERT"
TERRAIN_TUNDRA = "TERRAIN_TUNDRA"
TERRAIN_SNOW = "TERRAIN_SNOW"
TERRAIN_COAST = "TERRAIN_COAST"
TERRAIN_OCEAN = "TERRAIN_OCEAN"
TERRAIN_PEAK = "TERRAIN_PEAK"
TERRAIN_HILL = "TERRAIN_HILL"  

FEATURE_ICE = "FEATURE_ICE"
FEATURE_JUNGLE = "FEATURE_JUNGLE"
FEATURE_OASIS = "FEATURE_OASIS"
FEATURE_FLOOD_PLAINS = "FEATURE_FLOOD_PLAINS"
FEATURE_FOREST = "FEATURE_FOREST"

BONUS_DEER = "BONUS_DEER"
BONUS_FUR = "BONUS_FUR"
BONUS_IVORY = "BONUS_IVORY"
BONUS_CORN = "BONUS_CORN"
BONUS_RICE = "BONUS_RICE"
BONUS_WHEAT = "BONUS_WHEAT"
BONUS_CLAM = "BONUS_CLAM"
BONUS_CRAB = "BONUS_CRAB"
BONUS_FISH = "BONUS_FISH"
BONUS_GEMS = "BONUS_GEMS"
BONUS_GOLD = "BONUS_GOLD"
BONUS_SILVER = "BONUS_SILVER"
BONUS_MARBLE = "BONUS_MARBLE"
BONUS_STONE = "BONUS_STONE"
BONUS_BANANA = "BONUS_BANANA"
BONUS_COW = "BONUS_COW"
BONUS_PIG = "BONUS_PIG"
BONUS_HORSE = "BONUS_HORSE"
BONUS_SHEEP = "BONUS_SHEEP"
BONUS_DYE = "BONUS_DYE"
BONUS_INCENSE = "BONUS_INCENSE"
BONUS_SILK = "BONUS_SILK"
BONUS_SPICES = "BONUS_SPICES"
BONUS_SUGAR = "BONUS_SUGAR"
BONUS_WINE = "BONUS_WINE"
BONUS_WHALE = "BONUS_WHALE"

TRAIT_PHILOSOPHICAL = "TRAIT_PHILOSOPHICAL"
TRAIT_AGGRESSIVE = "TRAIT_AGGRESSIVE"
TRAIT_SPIRITUAL = "TRAIT_SPIRITUAL"
TRAIT_EXPANSIVE = "TRAIT_EXPANSIVE"
TRAIT_INDUSTRIOUS = "TRAIT_INDUSTRIOUS"
TRAIT_CREATIVE = "TRAIT_CREATIVE"
TRAIT_FINANCIAL = "TRAIT_FINANCIAL"
TRAIT_ORGANIZED = "TRAIT_ORGANIZED"
TRAIT_CHARISMATIC = "TECH_MYSTICISM"
TRAIT_IMPERIALIST = "TECH_THE_WHEEL"
TRAIT_PROTECTIVE = "TECH_HUNTING"

# Change the value to increase or decrease the number of starting technologies
# each civilization will have each time a new game is started.
# Default value is 2
g_iStartingTechnologyCount = 2

# Change the value to increase or decrease the range of plots that influence 
# the starting technologies for each civilization.
# Default value is 2
g_iPlotInfluenceRange = 2

# Change the value to false to exclude the plot type when calculating the 
# localized starting technologies for a civilization.
# Default value is true
g_bPlotTypeImpactsCalculation = true

# Change the value to increase or decrease the impact that the plot type of a
# plot should have when calculating the localized starting technologies for a
# civilization.
# Default value is 1
g_iPlotTypeWeight = 1

# Change the value to false to exclude the terrain type when calculating the 
# localized starting technologies for a civilization.
# Default value is true
g_bTerrainTypeImpactsCalculation = true

# Change the value to increase or decrease the impact that the terrain type of 
# a plot should have when calculating the localized starting technologies for a
# civilization.
# Default value is 1
g_iTerrainTypeWeight = 1

# Change the value to false to exclude the feature type when calculating the 
# localized starting technologies for a civilization.
# Default value is true
g_bFeatureTypeImpactsCalculation = true

# Change the value to increase or decrease the impact that the feature type of 
# a plot should have when calculating the localized starting technologies for a
# civilization.
# Default value is 1
g_iFeatureTypeWeight = 1

# Change the value to false to exclude the bonus type when calculating the 
# localized starting technologies for a civilization.
# Default value is true
g_bBonusTypeImpactsCalculation = true

# Change the value to increase or decrease the impact that the bonus type of a
# plot should have when calculating the localized starting technologies for a
# civilization.
# Default value is 1
g_iBonusTypeWeight = 1

# Change the value to false to exclude the leader trait when calculating the 
# localized starting technologies for a civilization.
# Default value is true
g_bLeaderTraitImpactsCalculation = true

# Change the value to increase or decrease the impact that the leader trait 
# should have when calculating the localized starting technologies for a
# civilization.
# Default value is 2
g_iLeaderTraitWeight = 2

# Add or remove technologies that should be considered as starting technologies
# for the civilizations in the game. These same technologies must be also be 
# included in one of the lists for plots types, terrain types, feature types 
# and bonus types.
# Default value is [TECH_FISHING, TECH_AGRICULTURE, TECH_THE_WHEEL, TECH_MYSTICISM, TECH_HUNTING, TECH_MINING]
g_startingTechList = [TECH_FISHING, TECH_AGRICULTURE, TECH_THE_WHEEL, TECH_MYSTICISM, TECH_HUNTING, TECH_MINING]

# Change the value to false to exclude plots with rivers when calculating the 
# localized starting technologies for a civilization.
# Default value is true
g_bPlotsWithRiversImpactsCalculation = true

# Add or remove technologies that should be considered when calculating the 
# starting technologies for civilization that start near plots with rivers.
# Default value is TECH_FISHING, TECH_AGRICULTURE
g_riverTechnologyMappings = [TECH_FISHING, TECH_AGRICULTURE]

# Change the value to increase or decrease the impact that the plots with 
# rivers should have when calculating the localized starting technologies for a
# civilization.
# Default value is 1
g_iRiverTechnologyMappingWeight = 1

# The following lines are the plot type to technology mappings that are used to
# determine the starting technologies for civilizations.
# Default values are:
#		PLOT_HILLS = TECH_MINING, TECH_MINING, TECH_MINING, TECH_MYSTICISM
#		PLOT_LAND = TECH_AGRICULTURE, TECH_THE_WHEEL, TECH_HUNTING
#		PLOT_OCEAN = TECH_FISHING, TECH_FISHING
#		PLOT_PEAK = TECH_MINING, TECH_MINING
g_PlotTypeTechMapping = {
							PLOT_HILLS	:	[TECH_MINING, TECH_MINING, TECH_MINING, TECH_MYSTICISM],
							PLOT_LAND	:	[TECH_AGRICULTURE, TECH_THE_WHEEL, TECH_HUNTING],
							PLOT_OCEAN	:	[TECH_FISHING, TECH_FISHING],
							PLOT_PEAK	:	[TECH_MINING, TECH_MINING],
						}

# The following lines are the terrain type to technology mappings that are used 
# to determine the starting technologies for civilizations.
# Default values are:
#		TERRAIN_GRASS = TECH_HUNTING, TECH_THE_WHEEL
#		TERRAIN_PLAINS = TECH_HUNTING, TECH_THE_WHEEL, TECH_MINING
#		TERRAIN_DESERT = TECH_MYSTICISM, TECH_MINING
#		TERRAIN_TUNDRA = TECH_MYSTICISM
#		TERRAIN_SNOW = TECH_MYSTICISM
#		TERRAIN_COAST = TECH_FISHING, TECH_MYSTICISM
#		TERRAIN_OCEAN = TECH_FISHING, TECH_MYSTICISM
#		TERRAIN_HILL = TECH_HUNTING, TECH_MINING, TECH_MINING, TECH_MYSTICISM  
#		TERRAIN_PEAK = TECH_MINING, TECH_MINING
g_TerrainTypeTechMapping =	{
							TERRAIN_GRASS : [TECH_HUNTING, TECH_THE_WHEEL],
							TERRAIN_PLAINS : [TECH_HUNTING, TECH_THE_WHEEL, TECH_MINING],
							TERRAIN_DESERT : [TECH_MYSTICISM, TECH_MINING],
							TERRAIN_TUNDRA : [TECH_MYSTICISM],
							TERRAIN_SNOW : [TECH_MYSTICISM],
							TERRAIN_COAST : [TECH_FISHING, TECH_MYSTICISM],
							TERRAIN_OCEAN : [TECH_FISHING, TECH_MYSTICISM],
							TERRAIN_HILL : [TECH_HUNTING, TECH_MINING, TECH_MINING, TECH_MYSTICISM],  
							TERRAIN_PEAK : [TECH_MINING, TECH_MINING],
						}				
						
# The following lines are the feature type to technology mappings that are used 
# to determine the starting technologies for civilizations.
# Default values are:
#		FEATURE_ICE = TECH_MYSTICISM
#		FEATURE_JUNGLE = TECH_MINING
#		FEATURE_OASIS = TECH_MYSTICISM
#		FEATURE_FLOOD_PLAINS = TECH_AGRICULTURE
#		FEATURE_FOREST = TECH_MYSTICISM						
g_FeatureTypeTechMapping = {
							FEATURE_ICE : [TECH_MYSTICISM],
							FEATURE_JUNGLE : [TECH_MINING],
							FEATURE_OASIS : [TECH_MYSTICISM],
							FEATURE_FLOOD_PLAINS : [TECH_MYSTICISM],
							FEATURE_FOREST : [TECH_MYSTICISM],
  						}

# The following lines are the bonus type to technology mappings that are used to
# determine the starting technologies for civilizations.
# Default values are:
#		BONUS_DEER = TECH_HUNTING
#		BONUS_FUR = TECH_HUNTING
#		BONUS_IVORY = TECH_HUNTING
#		BONUS_CORN = TECH_AGRICULTURE
#		BONUS_RICE = TECH_AGRICULTURE
#		BONUS_WHEAT = TECH_AGRICULTURE
#		BONUS_CLAM = TECH_FISHING
#		BONUS_CRAB = TECH_FISHING
#		BONUS_FISH = TECH_FISHING
#		BONUS_GEMS = TECH_MINING, TECH_MINING, TECH_MINING
#		BONUS_GOLD = TECH_MINING, TECH_MINING, TECH_MINING
#		BONUS_SILVER = TECH_MINING, TECH_MINING, TECH_MINING
#		BONUS_MARBLE = TECH_MINING, TECH_MINING, TECH_MINING
#		BONUS_STONE = TECH_MINING, TECH_MINING, TECH_MINING
#		BONUS_BANANA = TECH_MYSTICISM
#		BONUS_COW = TECH_AGRICULTURE
#		BONUS_HORSE = TECH_AGRICULTURE
#		BONUS_SHEEP = TECH_AGRICULTURE
#		BONUS_DYE = TECH_MYSTICISM
#		BONUS_INCENSE = TECH_MYSTICISM
#		BONUS_SILK = TECH_MYSTICISM
#		BONUS_SPICES = TECH_MYSTICISM
#		BONUS_SUGAR = TECH_MYSTICISM
#		BONUS_WINE = TECH_MYSTICISM
#		BONUS_WHALE = TECH_FISHING
g_BonusTypeTechMapping = {
							BONUS_DEER : [TECH_HUNTING],
							BONUS_FUR : [TECH_HUNTING],
							BONUS_IVORY : [TECH_HUNTING],
							BONUS_CORN : [TECH_AGRICULTURE],
							BONUS_RICE : [TECH_AGRICULTURE],
							BONUS_WHEAT : [TECH_AGRICULTURE],
							BONUS_CLAM : [TECH_FISHING],
							BONUS_CRAB : [TECH_FISHING],
							BONUS_FISH : [TECH_FISHING],
							BONUS_GEMS : [TECH_MINING, TECH_MINING, TECH_MINING],
							BONUS_GOLD : [TECH_MINING, TECH_MINING, TECH_MINING],
							BONUS_SILVER : [TECH_MINING, TECH_MINING, TECH_MINING],
							BONUS_MARBLE : [TECH_MINING, TECH_MINING, TECH_MINING],
							BONUS_STONE : [TECH_MINING, TECH_MINING, TECH_MINING],
							BONUS_BANANA : [TECH_MYSTICISM],
							BONUS_COW : [TECH_AGRICULTURE],
							BONUS_HORSE : [TECH_AGRICULTURE],
							BONUS_SHEEP : [TECH_AGRICULTURE],
							BONUS_DYE : [TECH_MYSTICISM],
							BONUS_INCENSE : [TECH_MYSTICISM],
							BONUS_SILK : [TECH_MYSTICISM],
							BONUS_SPICES : [TECH_MYSTICISM],
							BONUS_SUGAR : [TECH_MYSTICISM],
							BONUS_WINE : [TECH_MYSTICISM],
							BONUS_WHALE : [TECH_FISHING],
						}
						
# The following lines are the leader trait to technology mappings that are used 
# to determine the starting technologies for civilizations.
# Default values are:
#		TRAIT_PHILOSOPHICAL = TECH_MYSTICISM
#		TRAIT_AGGRESSIVE = TECH_THE_WHEEL
#		TRAIT_SPIRITUAL = TECH_MYSTICISM, TECH_MYSTICISM, TECH_MYSTICISM, TECH_MYSTICISM, TECH_MYSTICISM
#		TRAIT_EXPANSIVE = TECH_THE_WHEEL
#		TRAIT_INDUSTRIOUS = TECH_MINING
#		TRAIT_CREATIVE = TECH_FISHING
#		TRAIT_FINANCIAL = TECH_AGRICULTURE
#		TRAIT_ORGANIZED = TECH_HUNTING, TECH_AGRICULTURE
#		TRAIT_CHARISMATIC = TECH_MYSTICISM
#		TRAIT_IMPERIALIST = TECH_THE_WHEEL
#		TRAIT_PROTECTIVE = TECH_HUNTING
g_LeaderTraitTechMapping = {
								TRAIT_PHILOSOPHICAL : [TECH_MYSTICISM],
								TRAIT_AGGRESSIVE : [TECH_THE_WHEEL],
								TRAIT_SPIRITUAL : [TECH_MYSTICISM, TECH_MYSTICISM, TECH_MYSTICISM, TECH_MYSTICISM, TECH_MYSTICISM],
								TRAIT_EXPANSIVE : [TECH_THE_WHEEL],
								TRAIT_INDUSTRIOUS : [TECH_MINING],
								TRAIT_CREATIVE : [TECH_FISHING],
								TRAIT_FINANCIAL : [TECH_AGRICULTURE],
								TRAIT_ORGANIZED : [TECH_HUNTING, TECH_AGRICULTURE],
								TRAIT_CHARISMATIC : [TECH_MYSTICISM],
								TRAIT_IMPERIALIST : [TECH_THE_WHEEL],
								TRAIT_PROTECTIVE : [TECH_HUNTING],
							}
							
							
# Set to true to print out debug messages in the logs
g_bDebug = false

LOG_ONLY = 1
SCREEN_ONLY = 2
LOG_AND_SCREEN = 3

# Change to redirect the debug messages to the logs, screen or both.
g_iDebugMessageTarget = LOG_ONLY


def loadConfigurationData():
	global g_iStartingTechnologyCount
	global g_iPlotInfluenceRange
	global g_bPlotTypeImpactsCalculation
	global g_iPlotTypeWeight
	global g_bTerrainTypeImpactsCalculation
	global g_iTerrainTypeWeight
	global g_bFeatureTypeImpactsCalculation
	global g_iFeatureTypeWeight
	global g_bBonusTypeImpactsCalculation
	global g_iBonusTypeWeight
	global g_bLeaderTraitImpactsCalculation
	global g_iLeaderTraitWeight
	global g_startingTechList
	global g_bPlotsWithRiversImpactsCalculation
	global g_riverTechnologyMappings
	global g_iRiverTechnologyMappingWeight
	global g_PlotTypeTechMapping
	global g_TerrainTypeTechMapping
	global g_FeatureTypeTechMapping 
	global g_BonusTypeTechMapping	
	global g_LeaderTraitTechMapping
	
	config = CvConfigParser.CvConfigParser("Localized Starting Techs Mod Config.ini")

	if(config != None):

		g_iStartingTechnologyCount = config.getint("Localized Starting Techs Mod", "Starting Technology Count", 2)
		
		g_iPlotInfluenceRange = config.getint("Localized Starting Techs Mod", "Plot Influence Range", 2)

		g_bPlotTypeImpactsCalculation = config.getboolean("Localized Starting Techs Mod", "Plot Type Impacts Calculation", true)

		g_iPlotTypeWeight = config.getint("Localized Starting Techs Mod", "Plot Type Weight", 1)

		g_bTerrainTypeImpactsCalculation = config.getboolean("Localized Starting Techs Mod", "Terrain Type Impacts Calculation", true)

		g_iTerrainTypeWeight = config.getint("Localized Starting Techs Mod", "Terrain Type Weight", 1)

		g_bFeatureTypeImpactsCalculation = config.getboolean("Localized Starting Techs Mod", "Feature Type Impacts Calculation", true)

		g_iFeatureTypeWeight = config.getint("Localized Starting Techs Mod", "Feature Type Weight", 1)

		g_bBonusTypeImpactsCalculation = config.getboolean("Localized Starting Techs Mod", "Bonus Type Impacts Calculation", true)

		g_iBonusTypeWeight = config.getint("Localized Starting Techs Mod", "Bonus Type Weight", 1)

		g_bLeaderTraitImpactsCalculation = config.getboolean("Localized Starting Techs Mod", "Leader Trait Impacts Calculation", true)

		g_iLeaderTraitWeight = config.getint("Localized Starting Techs Mod", "Leader Trait Weight", 2)

		g_startingTechList = config.get("Localized Starting Techs Mod", "Starting Tech List", "TECH_FISHING, TECH_AGRICULTURE, TECH_THE_WHEEL, TECH_MYSTICISM, TECH_HUNTING, TECH_MINING")

		g_startingTechList = g_startingTechList.split(", ")

		g_bPlotsWithRiversImpactsCalculation = config.getboolean("Localized Starting Techs Mod", "Plots With Rivers Impacts Calculation", true)
			
		g_riverTechnologyMappings = config.get("Localized Starting Techs Mod", "Starting Tech List", "TECH_FISHING, TECH_FISHING, TECH_FISHING, TECH_AGRICULTURE")

		g_riverTechnologyMappings = g_riverTechnologyMappings.split(", ")

		g_iRiverTechnologyMappingWeight = config.getint("Localized Starting Techs Mod", "River Technology Mapping Weight", 1)


		configValues = config.items("Plot Type Technology Mappings")

		for tup in configValues:
			g_PlotTypeTechMapping[g_plotMap[tup[0].upper()]] = tup[1].split(", ")

		configValues = config.items("Terrain Type Technology Mappings")

		for tup in configValues:
			g_TerrainTypeTechMapping[tup[0].upper()] = tup[1].split(", ")
			
		configValues = config.items("Feature Type Technology Mappings")

		for tup in configValues:
			g_FeatureTypeTechMapping[tup[0].upper()] = tup[1].split(", ")
			
		configValues = config.items("Bonus Type Technology Mappings")

		for tup in configValues:
			g_BonusTypeTechMapping[tup[0].upper()] = tup[1].split(", ")
			
		configValues = config.items("Leader Trait Technology Mappings")

		for tup in configValues:
			g_LeaderTraitTechMapping[tup[0].upper()] = tup[1].split(", ")
	
			
			
class CvLocalizedStartingTechsEventManager:
	
	def __init__(self, eventManager):

		eventManager.addEventHandler("GameStart", self.onGameStart, False)
		#eventManager.addEventHandler("mouseEvent", self.onMouseEvent)
		eventManager.addEventHandler("windowActivation", self.onWindowActivation)


	def onWindowActivation(self, argsList):
		'Called when the game window activates or deactivates'
		bActive = argsList[0]
		if(bActive):
			loadConfigurationData()
		
	def onMouseEvent(self, argsList):
		
		loadConfigurationData()
		
		self.clearAllStartingTechs()

		self.assignStartingTechs()
		
		
	def onGameStart(self, argsList):
		loadConfigurationData()
		
		self.clearAllStartingTechs()

		self.assignStartingTechs()
	

	# Assigns the new starting technologies for the players in the game.
	def assignStartingTechs(self):

		# Get all of the players in the game
		playerList = PyGame.getCivPlayerList()
		
		# Go through all the players in the game
		for objPlayer in playerList:

			# Get the starting plot for the current player
			objPlot = objPlayer.CyGet().getStartingPlot()

			# Debug Code - Start
			if(g_bDebug):
				CvUtil.pyPrint("\n%s starting plot:(%s %s)" %(objPlayer.getName(), objPlot.getX(), objPlot.getY()))
			# Debug Code - End
			
			# Get the plot list for the configured radius
			plotList = self.getPlotList(objPlot, g_iPlotInfluenceRange)

			# Get the player's new starting technologies
			startingTechnologyList = self.getStartingTechnologies(plotList, objPlayer.CyGet())

			# Get the player's civilization information
			civInfo = objPlayer.getCivilizationInfo()

			# Go through all of the new starting technologies
			for strTechnology in startingTechnologyList:

				# Get the technology ID
				iTechType = gc.getInfoTypeForString(strTechnology)
								
				# Give the player the new technology								
				objPlayer.getTeam().setHasTech(iTechType, true, objPlayer.getID(), false, false)
				
				# Set the new technology as one of the free technologies
				civInfo.setCivilizationFreeTech(iTechType, true)

	
	# Returns the list of starting technologies derived from the plot list and 
	# player passed in
	def getStartingTechnologies(self, plotList, objPlayer):

		# Return an empty list if the plot list passed in is invalid
		if(plotList == None):
			return []
						
		# Return an empty list if the plot list passed in is invalid
		if(len(plotList) == 0):
			return []

		# Return an empty list if the player passed in is invalid
		if(objPlayer == None):
			return []
			
		# Return an empty list if the player passed in is invalid
		if(objPlayer.isNone()):
			return []

		# Get the player's team
		iTeam = objPlayer.getTeam()

		# Get the list of the player's leader traits
		leaderTraitList = self.getLeaderTraits(objPlayer)
								
		startingTechnologyHash = {}

		# Initialize the starting technology hash		
		for tech in g_startingTechList:
			startingTechnologyHash[tech] = 0
		
		# Go through all the plots in the plot list
		for objPlot in plotList:

			# Update the startingTechnologyHash if the current plot has a river
			# and plots with rivers impact the starting technology calculations
			if(objPlot.isRiver() and g_bPlotsWithRiversImpactsCalculation):
				startingTechnologyHash = self.calculatePlotWithRiverImpact(objPlot, startingTechnologyHash)

			# Update the startingTechnologyHash if the current plot is mapped 
			# to a technology and plot types impact the starting technology 
			# calculations
			if(g_PlotTypeTechMapping.has_key(objPlot.getPlotType()) and g_bPlotTypeImpactsCalculation):
				startingTechnologyHash = self.calculatePlotTypeImpact(objPlot, startingTechnologyHash)

			# Get the plot terrain info
			objTerrainInfo = gc.getTerrainInfo(objPlot.getFeatureType())

			# Update the startingTechnologyHash if the terrain type is mapped 
			# to a technology and terrain types impact the starting technology 
			# calculations			
			if(objTerrainInfo != None and g_TerrainTypeTechMapping.has_key(objTerrainInfo.getType()) and g_bTerrainTypeImpactsCalculation):
				startingTechnologyHash = self.calculateTerrainTypeImpact(objPlot, startingTechnologyHash)

			# Get the plot feature info
			objFeatureInfo = gc.getFeatureInfo(objPlot.getFeatureType())

			# Update the startingTechnologyHash if the feature type is mapped 
			# to a technology and feature types impact the starting technology 
			# calculations			
			if(objFeatureInfo != None and g_FeatureTypeTechMapping.has_key(objFeatureInfo.getType()) and g_bFeatureTypeImpactsCalculation):
				startingTechnologyHash = self.calculateFeatureTypeImpact(objPlot, startingTechnologyHash)

			# Get the plot bonus info
			objBonusInfo = gc.getBonusInfo(objPlot.getBonusType(iTeam))

			# Update the startingTechnologyHash if the bonus type is mapped 
			# to a technology and bonus types impact the starting technology 
			# calculations			
			if(objBonusInfo != None and g_BonusTypeTechMapping.has_key(objBonusInfo.getType()) and g_bBonusTypeImpactsCalculation):
				startingTechnologyHash = self.calculateBonusTypeImpact(objPlot, startingTechnologyHash, iTeam)


		# Update the startingTechnologyHash if the leader traits impact the 
		# starting technology calculations			
		if(g_bLeaderTraitImpactsCalculation):
			
			# Go through the list of leader traits and get their impact to the 
			# starting technologies
			for objTraitInfo in leaderTraitList:
				startingTechnologyHash = self.calculateLeaderTraitImpact(objTraitInfo, startingTechnologyHash)
					
		reverseStartingTechnologyHash = {}
		
		# Go through all the technologies retrieved
		for strTech in startingTechnologyHash:
		
			# Add the technology to the reverseStartingTechnologyHash
			reverseStartingTechnologyHash[startingTechnologyHash[strTech]] = strTech

			# Debug code - Start
			if(g_bDebug):
				CvUtil.pyPrint("%s %s" %(strTech, startingTechnologyHash[strTech]))
			# Debug code - End

		# Get the keys for the reverseStartingTechnologyHash	
		technologyList = reverseStartingTechnologyHash.keys()
		
		# Sort the technology list
		technologyList.sort()
		
		# Reverse the order of the technology list
		technologyList.reverse()
		
		# Get the slice of the technology list for the configured start 
		# technology amount
		technologyList = technologyList[0:g_iStartingTechnologyCount]
		
		# Go through all the technologies in the technology list
		for i in range(len(technologyList)):
		
			technologyList[i] = reverseStartingTechnologyHash[technologyList[i]]

			# Debug code - Start
			if(g_bDebug):
				CvUtil.pyPrint("Top pick %s: %s" %(i, technologyList[i]))
			# Debug code - End
		
		return technologyList
		
	
	# Calculate the plot with river impact to the starting technology 
	# calculations
	def calculatePlotWithRiverImpact(self, objPlot, startingTechnologyHash):

		# Go through all the tech types in the list
		for techType in g_riverTechnologyMappings:

			if(not startingTechnologyHash.has_key(techType)):
				continue
				
			startingTechnologyHash[techType] += g_iRiverTechnologyMappingWeight
	
		return startingTechnologyHash
		
		
	# Calculate the plot type impact to the starting technology calculations
	def calculatePlotTypeImpact(self, objPlot, startingTechnologyHash):

		# Get the plot type to tech mapping
		plotTechList = g_PlotTypeTechMapping[objPlot.getPlotType()]
		
		# Go through all the tech types in the list
		for techType in plotTechList:

			if(not startingTechnologyHash.has_key(techType)):
				continue
				
			startingTechnologyHash[techType] += g_iPlotTypeWeight
	
		return startingTechnologyHash
	
	
	# Calculate the terrain type impact to the starting technology calculations	
	def calculateTerrainTypeImpact(self, objPlot, startingTechnologyHash):

		# Get the terrain type to tech mapping
		terrainTechList = g_TerrainTypeTechMapping[gc.getTerrainInfo(objPlot.getTerrainType()).getType()]
			
		# Go through all the tech types in the list
		for techType in terrainTechList:
			if(not startingTechnologyHash.has_key(techType)):
				continue
				
			startingTechnologyHash[techType] += g_iTerrainTypeWeight
	
		return startingTechnologyHash
		
		
	# Calculate the feature type impact to the starting technology calculations	
	def calculateFeatureTypeImpact(self, objPlot, startingTechnologyHash):

		# Get the feature type to tech mapping
		featureTechList = g_FeatureTypeTechMapping[gc.getFeatureInfo(objPlot.getFeatureType()).getType()]
			
		# Go through all the tech types in the list
		for techType in featureTechList:
			if(not startingTechnologyHash.has_key(techType)):
				continue
				
			startingTechnologyHash[techType] += g_iFeatureTypeWeight
	
		return startingTechnologyHash		
	
		
	# Calculate the bonus type impact to the starting technology calculations	
	def calculateBonusTypeImpact(self, objPlot, startingTechnologyHash, iTeam):

		# Get the bonus type to tech mapping
		bonusTechList = g_BonusTypeTechMapping[gc.getBonusInfo(objPlot.getBonusType(iTeam)).getType()]
			
		# Go through all the tech types in the list
		for techType in bonusTechList:
			if(not startingTechnologyHash.has_key(techType)):
				continue
				
			startingTechnologyHash[techType] += g_iBonusTypeWeight
	
		return startingTechnologyHash
		

	# Calculate the leader trait impact to the starting technology calculations	
	def calculateLeaderTraitImpact(self, objTraitInfo, startingTechnologyHash):

		# Get the leader trait to tech mapping
		leaderTraitTechList = g_LeaderTraitTechMapping[objTraitInfo.getType()]
			
		# Go through all the tech types in the list
		for techType in leaderTraitTechList:
			if(not startingTechnologyHash.has_key(techType)):
				continue
				
			startingTechnologyHash[techType] += g_iLeaderTraitWeight
	
		return startingTechnologyHash				
	

	# Return the list of plots surrounding the plot passed in for the range 
	# passed in	
	def getPlotList(self, objPlot, iRange):
	
		# Return an empty list if the plot passed in is invalid
		if(objPlot == None):
			return []

		# Return an empty list if the plot passed in is invalid
		if(objPlot.isNone()):
			return []
			
		# Return an empty list if the ranged passed in is invalid
		if(iRange <= 0):
			return []

		# Get the starting location for the plot passed in
		iStartX = objPlot.getX()
		iStartY = objPlot.getY()

		# Get the map extents for the plot passed in
		iMinX = iStartX-iRange
		iMaxX = iStartX+iRange+1
		iMinY = iStartY-iRange
		iMaxY = iStartY+iRange+1

		plotList = []
		
		objMap = CyMap()
		
		# Go through the horizontal extents
		for currentX in range(iMinX, iMaxX):

		# Go through the vertical extents
			for currentY in range(iMinY, iMaxY):
	
				# Continue if the current plot coordinates are invalid
				if(not objMap.isPlot(currentX, currentY)):
					continue

				# Append the current plot to the plot list					
				plotList.append(objMap.plot(currentX, currentY))

		# Debug Code - Start
		if(g_bDebug):
			CyInterface().addImmediateMessage("(%s, %s) (%s, %s) %s" %(iMinX, iMinY, iMaxX, iMaxY, len(plotList)),"")
			
			strCoor = ""
			
			for plot in plotList:
				strCoor += "(%s,%s) " %(plot.getX(), plot.getY())
			
			CyInterface().addImmediateMessage(strCoor,"")
		# Debug Code - End

		return plotList
	
	
	# Removes all of the starting technologies for the players in the game
	def clearAllStartingTechs(self):

		# Get the players in the game
		playerList = PyGame.getCivPlayerList()
		
		# Go through all the players in the game
		for player in playerList:
		
			# Get the civilization information for the current player
			civInfo = player.getCivilizationInfo()
			
			# Get the current players free tech count
			iFreeTechCount = civInfo.getNumCivilizationFreeTechs()
			
			# Continue processing while the civilizations free tech count is 
			# above 0
			while(iFreeTechCount > 0):
	
				# Get the civilization's info first free tech
				tech = civInfo.getCivilizationFreeTech(0)
				
				# Remove the current technology from the player
				player.getTeam().setHasTech(gc.getInfoTypeForString(tech.getType()), false, player.getID(), false, false)
				
				# Remove the current technology from the player as a starting technology
				civInfo.setCivilizationFreeTech(gc.getInfoTypeForString(tech.getType()), false)

				iFreeTechCount = civInfo.getNumCivilizationFreeTechs()
					

	# Returns the list of leader traits for the player passed in
	def getLeaderTraits(self, objPlayer):

		# Return an empty list if the player passed in is invalid
		if(objPlayer == None):
			return []
			
		# Return an empty list if the player passed in is invalid
		if(objPlayer.isNone()):
			return []

		leaderTraitList = []
		
		# Get the ID of the leader type for the player
		iLeaderHeadType = objPlayer.getLeaderType()
		
		# Get the leader type for the player
		objLeaderHeadInfo = gc.getLeaderHeadInfo(iLeaderHeadType)
		
		strLeaderTraits = ""
					
		# Go through all of the leader traits in the game
		for i in range(gc.getNumTraitInfos()):
		
			# If the player has the current trait then append it to the leader 
			# trait list.
			if(objLeaderHeadInfo.hasTrait(i)):
				leaderTraitList.append(gc.getTraitInfo(i))
				strLeaderTraits += "%s|" %(gc.getTraitInfo(i).getType())

		# Debug code - Start
		if(g_bDebug):
			CvUtil.pyPrint("Leader Traits for %s are: |%s" %(objPlayer.getName(),strLeaderTraits))
			
		# Debug code - End

		return leaderTraitList
		